<?php

/**
 * @file
 *   Example drush command.
 *
 *   To run this *fun* command, execute `sudo drush --include=./examples mmas`
 *   from within your drush directory.
 *
 *   See `drush topic docs-commands` for more information about command authoring.
 *
 *   You can copy this file to any of the following
 *     1. A .drush folder in your HOME folder.
 *     2. Anywhere in a folder tree below an active module on your site.
 *     3. /usr/share/drush/commands (configurable)
 *     4. In an arbitrary folder specified with the --include option.
 *     5. Drupal's /drush or /sites/all/drush folders.
 */

/**
 * Implementation of hook_drush_command().
 */
function monkey_drush_command() {
  $items = array();

  // The hard-working 'monkey-check' command
  $items['monkey-check'] = array(
    'description' => "Hard-working Monkey do some checks of your running Drupal site.",
    'arguments' => array(
      'check' => 'The type of the check (fields, etc.). Defaults to fields.',
      'entity_type' => 'The entity type to check.',
      'bundle' => 'The bundle to check.'
    ),
    'options' => array(
      'fields' => array(
        'description' => 'Comma delimited list of field machine names.',
        'example-value' => 'field_body,field_tags',
      ),
    ),
    'examples' => array(
      'drush monkey-checks fields --fields=field_body ' => 'Make a terrible-tasting sandwich that is lacking in pickles.',
    ),
    'bootstrap' => DRUSH_BOOTSTRAP_DRUPAL_FULL, // Full bootstrap for required Drupal functions.
  );

  // The hard-working 'monkey-fix' command where our monkey try to fix fields, etc.
  $items['monkey-fix'] = array(
    'description' => "Hard-working Monkey fix something in your running Drupal site.",
    'arguments' => array(
      'fix' => 'The type of the fix (fields, etc.). Defaults to fields.',
      'entity_type' => 'The entity type to check.',
      'bundle' => 'The bundle to check.'
    ),
    'options' => array(
      'fields' => array(
        'description' => 'Comma delimited list of field machine names.',
        'example-value' => 'field_body,field_tags',
      ),
      'fast' => array(
        'description' => 'With this option hard-working monkey do his jobs faster (experimental).',
      ),
      'with-batch' => array(
        'description' => 'With this option hard-working monkey do his jobs as batch operations.',
      ),
      'batch-items' => array(
        'description' => 'With this option you can tell hard-working monkey how many items during a batch operation should being fixed.',
      ),
    ),
    'examples' => array(
      'drush monkey-fix fields node article --fields=field_xyz' => 'Monkey try to fix the field_xyz field for article node entities.',
    ),
    'bootstrap' => DRUSH_BOOTSTRAP_DRUPAL_FULL, // Full bootstrap for required Drupal functions.
  );

  return $items;
}

/**
 * Implementation of hook_drush_help().
 */
function monkey_drush_help($section) {
  switch ($section) {
    case 'drush:monkey-check':
      return dt("This command will make some checks by hard-working monkey.");
    case 'drush:monkey-fix':
      return dt("This command will make some fixes by hard-working monkey.");
  }
}

/**
 * Implementation of hook_drush_command() for monkey-check command.
 */
function drush_monkey_check($check = 'fields', $entity_type = 'node', $bundle = 'page') {

  $instances = _monkey_get_instances($entity_type, $bundle);
  $fields = array_keys($instances);

  $msg = dt('Monkey checks following !check: !fields.',
    array('!check' => $check, '!fields' => implode(', ', $fields))
  );
  drush_print($msg);

  $query = new EntityFieldQuery();
  $query->entityCondition('entity_type', $entity_type)
    ->entityCondition('bundle', $bundle);
  $result = $query->execute();

  $msg = dt('Monkey checks for !count !bundle !entity_type(s).',
    array(
      '!count' => count($result[$entity_type]),
      '!bundle' => $bundle,
      '!entity_type' => $entity_type,
    )
  );
  drush_print("\n" . $msg . "\n");

  foreach ($fields as $field) {
    $have_value_count = dt('Unknown field condition.');

    try {

      $query = new EntityFieldQuery();
      $query->entityCondition('entity_type', $entity_type)
        ->entityCondition('bundle', $bundle)
        ->fieldCondition($field, 'value', 'NULL', '!=');
      $result1 = $query->execute();

      $query = new EntityFieldQuery();
      $query->entityCondition('entity_type', $entity_type)
        ->entityCondition('bundle', $bundle)
        ->fieldCondition($field, 'value', 'NULL', '=');
      $result2 = $query->execute();

      $have_value_count = count($result1[$entity_type]) + count($result2[$entity_type]);

      $empty_default_count = dt('This field have no default value.');
      if (!empty($instances[$field]['default_value'])) {

        $values1 = !empty($result1[$entity_type]) ? $result1[$entity_type] : array();
        $values2 = !empty($result2[$entity_type]) ? $result2[$entity_type] : array();

        $entity_have_values = $values1 + $values2;

        if (is_array($entity_have_values)) {
          //Now get all the other entities, that aren't in the list you just retrieved
          $query = new EntityFieldQuery();
          $query->entityCondition('entity_type', $entity_type)
            ->entityCondition('bundle', $bundle)
            ->entityCondition('entity_id', array_keys($entity_have_values), 'NOT IN');
          $result = $query->execute();

          $empty_default_count = count($result[$entity_type]);
        }
      }
    }  catch (Exception $e) {
    }

    $checked_field = dt('Field: !field',
      array('!field' => $field)
    );
    $have_value = dt('Have value(s): !have_value_count',
      array('!have_value_count' => $have_value_count)
    );
    $empty_default = dt('Without default: !empty_default_count',
      array('!empty_default_count' => $empty_default_count)
    );
    drush_print($checked_field . "\n" . $have_value . "\n" . $empty_default . "\n");
  }
}

/**
 * Implementation of hook_drush_command() for monkey-fix command.
 */
function drush_monkey_fix($fix = 'fields', $entity_type = 'node', $bundle = 'page') {

  $instances = _monkey_get_instances($entity_type, $bundle);
  $fields = array_keys($instances);

  $msg = dt('Monkey try to fix following !fix: !fields.',
    array('!fix' => $fix, '!fields' => implode(', ', $fields))
  );
  drush_print($msg . "\n");

  foreach ($fields as $field) {
    $have_value_count = dt('Unknown field condition.');

    $fix_entitys = array();
    try {

      $query = new EntityFieldQuery();
      $query->entityCondition('entity_type', $entity_type)
        ->entityCondition('bundle', $bundle)
        ->fieldCondition($field, 'value', 'NULL', '!=');
      $result1 = $query->execute();

      $query = new EntityFieldQuery();
      $query->entityCondition('entity_type', $entity_type)
        ->entityCondition('bundle', $bundle)
        ->fieldCondition($field, 'value', 'NULL', '=');
      $result2 = $query->execute();

      $have_value_count = count($result1[$entity_type]) + count($result2[$entity_type]);

      $empty_default_count = dt('This field have no default value.');
      if (!empty($instances[$field]['default_value'])) {

        $values1 = !empty($result1[$entity_type]) ? $result1[$entity_type] : array();
        $values2 = !empty($result2[$entity_type]) ? $result2[$entity_type] : array();

        $entity_have_values = $values1 + $values2;

        if (is_array($entity_have_values)) {
          //Now get all the other entities, that aren't in the list you just retrieved
          $query = new EntityFieldQuery();
          $query->entityCondition('entity_type', $entity_type)
            ->entityCondition('bundle', $bundle)
            ->entityCondition('entity_id', array_keys($entity_have_values), 'NOT IN');
          $result = $query->execute();

          $empty_default_count = count($result[$entity_type]);

          if ($empty_default_count > 0) {
            $fix_entitys = $result[$entity_type];
          }
        }
      }
    }  catch (Exception $e) {
    }

    $checked_field = dt('Field: !field',
      array('!field' => $field)
    );
    $have_value = dt('Have value(s): !have_value_count',
      array('!have_value_count' => $have_value_count)
    );
    $empty_default = dt('Without default: !empty_default_count',
      array('!empty_default_count' => $empty_default_count)
    );
    drush_print($checked_field . "\n" . $have_value . "\n" . $empty_default . "\n");

    $fixed_count = 0;
    $with_batch = (bool) drush_get_option('with-batch');

    $fast = (bool) drush_get_option('fast');

    if ($with_batch) {

      $batch_items = (int) drush_get_option('batch-items');
      if ($batch_items == 0 || !is_int($batch_items) ) {
        $batch_items = 100;
      }

      $drush_options = array(
        'batch_items' => $batch_items,
        'fast' => $fast,
      );

      $batch = array(
        'title' => t('Monkey drush batch process to fix entities'),
        'operations' => array(
          array(
            'monkey_drush_batch_fix_operation',
            array(
              $entity_type,
              $fix_entitys,
              $field,
              $drush_options
            ),
          )
        ),
        'finished' => 'monkey_drush_batch_finished_batch',
      );

      batch_set($batch);
      $batch =& batch_get();
      $batch['progressive'] = FALSE;

      drush_backend_batch_process();
    }
    else {

      foreach($fix_entitys as $entity) {

        _monkey_fix_field($entity_type, $entity, $field, $fast);

        $fixed_count++;
        _monkey_drush_print_progress($fixed_count / count($fix_entitys));
      }

      $fixed = dt('Fixed: !fixed_count',
        array('!fixed_count' => $fixed_count)
      );
      drush_print($fixed . "\n");
    }
  }
}

/**
 *  Batch callback for fix operation.
 */
function monkey_drush_batch_fix_operation($entity_type, $fix_entitys, $field, $drush_options = array(), &$context) {

  // Init $context['sandbox'] variables.
  if (empty($context['sandbox'])) {

    $context['sandbox']['batch_items'] = $drush_options['batch_items'];
    $context['sandbox']['fast'] = $drush_options['fast'];

    $context['sandbox']['entitys'] = $fix_entitys;
    $context['sandbox']['max'] = count($fix_entitys);
    $context['sandbox']['progress'] = 0;
  }

  $limit = min($context['sandbox']['batch_items'], count($context['sandbox']['entitys']));

  for ($i = 0; $i < $limit; $i++) {

    $current_entity = array_shift($context['sandbox']['entitys']);

    if (_monkey_fix_field($entity_type, $current_entity, $field, $context['sandbox']['fast'])) {
      $context['results'][] = 'Successfully fixed ' . $field . ' of entity.';
    }
    else {
      $context['results'][] = 'Error while fixing ' . $field . ' of entity.';
    }

    $context['sandbox']['progress']++;
  }

  _monkey_drush_print_progress($context['sandbox']['progress'] / $context['sandbox']['max']);

  //check if batch is finished and update progress
  if ($context['sandbox']['progress'] != $context['sandbox']['max']) {
    $context['finished'] = $context['sandbox']['progress'] / $context['sandbox']['max'];
  }
}

function monkey_drush_batch_finished_batch($success, $results, $operations) {

  if ($success) {

    $fixed = dt('Fixed: !fixed_count',
      array('!fixed_count' => count($results))
    );
    drush_print($fixed . "\n");
  }
}

function _monkey_fix_field($entity_type, $entity, $field, $fast = FALSE) {

  if ($fast) {

    _monkey_fast_field_attach_insert($entity_type, $entity, $field);
  }
  else {

    _field_invoke_default('insert', $entity_type, $entity);
    field_attach_update($entity_type, $entity);
  }

  return TRUE;
}

/**
 * Helper function for faster field inserts based on Drupal Core field_attach_insert().
 *
 * @param $entity_type
 * @param $entity
 * @param $field_name
 */
function _monkey_fast_field_attach_insert($entity_type, $entity, $field_name) {
  _field_invoke_default('insert', $entity_type, $entity);
  _field_invoke('insert', $entity_type, $entity);

  list($id, $vid, $bundle) = entity_extract_ids($entity_type, $entity);

  // Let any module insert field data before the storage engine, accumulating
  // saved fields along the way.
  $skip_fields = array();
  foreach (module_implements('field_storage_pre_insert') as $module) {
    $function = $module . '_field_storage_pre_insert';
    $function($entity_type, $entity, $skip_fields);
  }

  // Collect the storage backends used by the remaining fields in the entities.
  $storages = array();
  $instance = field_info_instance($entity_type, $field_name, $bundle);

  //foreach (field_info_instances($entity_type, $bundle) as $instance) {
  $field = field_info_field_by_id($instance['field_id']);
  $field_id = $field['id'];
  $field_name = $field['field_name'];
  if (!empty($entity->$field_name)) {
    // Collect the storage backend if the field has not been written yet.
    if (!isset($skip_fields[$field_id])) {
      $storages[$field['storage']['type']][$field_id] = $field_id;
    }
  }
  //}

  // Field storage backends save any remaining unsaved fields.
  foreach ($storages as $storage => $fields) {
    $storage_info = field_info_storage_types($storage);
    module_invoke($storage_info['module'], 'field_storage_write', $entity_type, $entity, FIELD_STORAGE_INSERT, $fields);
  }

  // Let other modules act on inserting the entity.
  module_invoke_all('field_attach_insert', $entity_type, $entity);

  $entity_info = entity_get_info($entity_type);
}

function _monkey_get_instances($entity_type, $bundle) {

  // Read options with drush_get_option. Note that the options _must_
  // be documented in the $items structure for this command in the 'command' hook.
  // See `drush topic docs-commands` for more information.
  if ($fields = drush_get_option('fields')) {

    $fields = explode(',', $fields);

    $instances = array();
    foreach ($fields as $field) {

      $instance = field_info_instance($entity_type, $field, $bundle);
      if (!empty($instance)) {
        $instances[$field] = $instance;
      }
    }
  }
  else {
    $instances = field_info_instances($entity_type, $bundle);
  }

  return $instances;
}

/**
 * Provides progress bar.
 */
function _monkey_drush_print_progress($ratio) {
  $percentage = floor($ratio * 100) . '%';
  $columns = drush_get_context('DRUSH_COLUMNS', 80);
  // Subtract 8 characters for the percentage, brackets, spaces and arrow.
  $progress_columns = $columns - 8;
  // If ratio is 1 (complete), the > becomes a = to make a full bar.
  $arrow = ($ratio < 1) ? '>' : '=';
  // Print a new line if ratio is 1 (complete). Otherwise, use a CR.
  $line_ending = ($ratio < 1) ? "\r" : "\n";

  // Determine the current length of the progress string.
  $current_length = floor($ratio * $progress_columns);
  $progress_string = str_pad('', $current_length, '=');

  $output  = '[';
  $output .= $progress_string . $arrow;
  $output .= str_pad('', $progress_columns - $current_length);
  $output .= ']';
  $output .= str_pad('', 5 - strlen($percentage)) . $percentage;
  $output .= $line_ending;

  print $output;
}
